from collections.abc import Iterable
from dataclasses import dataclass, field
from pathlib import Path

import numpy as np
import xxhash
from colour.colorimetry.spectrum import MultiSpectralDistributions
from colour.hints import NDArray
from numpy import ndarray

from specio.common import SPDMeasurement
from specio.serialization.measurements import (
    spd_measurement_from_bytes,
    spd_measurement_to_proto,
)
from specio.serialization.protobuf import measurements_pb2

__author__ = "Tucker Downs"
__copyright__ = "Copyright 2022 Specio Developers"
__license__ = "MIT License - https://github.com/tjdcs/specio/blob/main/LICENSE.md"
__maintainer__ = "Tucker Downs"
__email__ = "tucker@tjdcs.dev"
__status__ = "Development"
__all__ = [
    "CSMF_Data",
    "CSMF_Metadata",
    "csmf_data_to_buffer",
    "save_csmf_file",
    "load_csmf_file",
]


@dataclass()
class CSMF_Metadata:
    """Several metadata strings relating to how a set of measurements was
    conducted.
    """

    notes: None | str = None
    author: None | str = None
    location: None | str = None
    software: None | str = "colour-specio"


@dataclass(kw_only=True)
class CSMF_Data:
    """List of measurements, test colors, measurement, order, and metadata.

    Fields
    -------
    _type_
        _description_
    """

    test_colors: ndarray
    order: Iterable[int]
    measurements: NDArray = field(
        default_factory=lambda: np.empty_like(prototype=SPDMeasurement)
    )
    metadata: CSMF_Metadata = field(default_factory=CSMF_Metadata)

    @property
    def shortname(self) -> str:
        """A short name, generated by hashing underlying data if metadata.notes
        is empty.

        Returns
        -------
        str
        """
        if self.metadata.notes is None or self.metadata.notes == "":
            spds = MultiSpectralDistributions([m.spd for m in self.measurements])
            return xxhash.xxh32_hexdigest(np.ascontiguousarray(spds.values).data)
        return self.metadata.notes

    def __repr__(self) -> str:
        return f"Measurement List - {self.shortname}"

    def __eq__(self, value: object) -> bool:
        if isinstance(value, CSMF_Data):
            return all(
                (
                    np.all(self.test_colors == value.test_colors),
                    np.all(self.order == value.order),
                    np.all(self.measurements == value.measurements),
                    self.metadata == value.metadata,
                )
            )
        return False


def csmf_data_to_buffer(
    ml: CSMF_Data,
) -> measurements_pb2.CSFM_File:
    pbuf = measurements_pb2.CSFM_File()

    for m in ml.measurements:
        m_pbuf = spd_measurement_to_proto(m)
        pbuf.spd_measurements.append(m_pbuf)

    if ml.metadata.notes:
        pbuf.notes = ml.metadata.notes

    if ml.metadata.author:
        pbuf.author = ml.metadata.author

    if ml.metadata.location:
        pbuf.location = ml.metadata.location

    if ml.metadata.software:
        pbuf.software = ml.metadata.software

    if ml.order is not None:
        pbuf.order[:] = ml.order

    if ml.test_colors is not None:
        testColors = np.asarray(ml.test_colors)

        if np.ptp(testColors) > 1 and np.all(np.modf(testColors)[0] < 1e-8):
            # If the test colors are ints, use the int field.
            for color in testColors:
                tc_buf = measurements_pb2.CSFM_File.TestColor()
                tc_buf.c[:] = [int(value) for value in color.tolist()]
                pbuf.test_colors.append(tc_buf)
        else:
            for color in testColors:
                tc_buf = measurements_pb2.CSFM_File.TestColor()
                tc_buf.f[:] = color
                pbuf.test_colors.append(tc_buf)
    return pbuf


def save_csmf_file(
    file: str | Path,
    ml: CSMF_Data,
) -> Path:
    buffer = csmf_data_to_buffer(ml)
    data_string = buffer.SerializeToString()

    if isinstance(file, str):
        file = Path(file)

    file = file.with_suffix(".csmf")

    with open(file=file, mode="wb") as f:
        f.write(data_string)

    return file


def load_csmf_file(file: str | Path, recompute: bool = False) -> CSMF_Data:
    """Load measurement list data from a file

    Parameters
    ----------
    file : str | Path
        The csmf file to read
    recompute : bool, optional
        Recomputes derived data from the spd data, i.e. XYZ, CCT, etc... Very
        slow for large files. by default False

    Returns
    -------
    Measurement_List
        Measurement Data
    """
    if isinstance(file, str):
        file = Path(file)
    file = file.with_suffix(".csmf")

    data_string: bytes
    with open(file, mode="rb") as f:
        data_string = f.read()

    pbuf = measurements_pb2.CSFM_File()
    pbuf.ParseFromString(data_string)

    measurements = []
    for mbuf in pbuf.spd_measurements:
        measurements.append(spd_measurement_from_bytes(mbuf, recompute=recompute))
    measurements = np.asarray(measurements)

    tcs = []
    for color in pbuf.test_colors:
        if len(color.f) == 0:
            tcs.append(color.c)
        else:
            tcs.append(color.f)
    tcs = np.array(tcs)

    return CSMF_Data(
        measurements=measurements,
        order=np.asarray(pbuf.order),
        test_colors=tcs,
        metadata=CSMF_Metadata(pbuf.notes, pbuf.author, pbuf.location, pbuf.software),
    )



"""
Define basic spectrometer interfaces
"""

import textwrap
from abc import ABC, abstractmethod
from ctypes import ArgumentError
from dataclasses import dataclass
from datetime import datetime
from functools import cached_property
from typing import Self, final

import numpy as np
from colour import sd_multi_leds
from colour.colorimetry.dominant import (
    colorimetric_purity,
    dominant_wavelength,
)
from colour.colorimetry.tristimulus_values import sd_to_XYZ
from colour.hints import ArrayLike
from colour.models.cie_xyy import XYZ_to_xy
from colour.temperature.ohno2013 import XYZ_to_CCT_Ohno2013

__author__ = "Tucker Downs"
__copyright__ = "Copyright 2022 Specio Developers"
__license__ = "MIT License - https://github.com/tjdcs/specio/blob/main/LICENSE.md"
__maintainer__ = "Tucker Downs"
__email__ = "tucker@tjdcs.dev"
__status__ = "Development"

__all__ = []


@dataclass
class RawColorimeterMeasurement:
    XYZ: np.ndarray
    exposure: float
    device_id: str


class ColorimeterMeasurement:
    @classmethod
    def FromRaw(cls, raw: RawColorimeterMeasurement) -> Self:
        return cls(raw.XYZ, raw.exposure, raw.device_id)

    def __init__(
        self,
        XYZ: ArrayLike,
        exposure: float,
        device_id: str,
        no_compute: bool = False,
    ):
        XYZ = np.asarray(XYZ)
        if XYZ.size != (3,):
            RuntimeError("XYZ must be size (3,)")

        self.XYZ = XYZ
        self.exposure = exposure
        self.device_id = device_id

        if not no_compute:
            cct = XYZ_to_CCT_Ohno2013(self.XYZ)
            self.cct: float = cct[0]
            self.duv: float = cct[1]

            self.xy = XYZ_to_xy(self.XYZ)
            self.dominant_wl: float = float(
                dominant_wavelength(self.xy, [1 / 3, 1 / 3])[0]
            )
            self.purity: float = colorimetric_purity(self.xy, (1 / 3, 1 / 3))  # type: ignore
            self.time = datetime.now().astimezone()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ColorimeterMeasurement):
            keys = [
                "XYZ",
                "exposure",
                "device_id",
                "cct",
                "duv",
                "xy",
                "dominant_wl",
                "purity",
                "time",
            ]
            bools = [np.all(getattr(self, k) == getattr(other, k)) for k in keys]
            return all(bools)
        return False

    def __str__(self) -> str:
        """
        Return printable string with interesting data for an observant reader.

        Returns
        -------
        str
        """
        return textwrap.dedent(
            f"""
            Colorimeter Measurement - {self.device_id}:
                time: {self.time}
                XYZ: {np.array2string(self.XYZ, formatter={'float_kind':lambda x: "%.2f" % x})}
                xy: {np.array2string(self.xy, formatter={'float_kind':lambda x: "%.4f" % x})}
                CCT: {self.cct:.0f} ± {self.duv:.5f}
                Dominant WL: {self.dominant_wl:.1f} @ {self.purity * 100:.1f}%
                Exposure: {self.exposure:.3f}
            """  # noqa: E501
        )


class Colorimeter(ABC):
    """
    Base class for spectroradiometers. Defines the minimum interface needed
    by implementing classes to collect measurement values from a spectrometer.
    """

    @property
    @abstractmethod
    def serial_number(self) -> str:
        """The serial number or id for the specific connected spectrometer

        Returns
        -------
        str
        """

        ...

    @property
    @abstractmethod
    def manufacturer(self) -> str:
        """The device manufacturer name

        Returns
        -------
        str
        """

        ...

    @cached_property
    @abstractmethod
    def model(self) -> str:
        """The model name or identifier for the device

        Returns
        -------
        str
        """

        ...

    @abstractmethod
    def _raw_measure(self) -> RawColorimeterMeasurement:
        """Conduct remote triggering and RPC / serial interaction to collect
        the raw spectrometer data from the device.

        Returns
        -------
        RawMeasurement
            A simple dataclass with the required parameters to produce fully
            defined :class:`specio.spectrometers.common.Measurement`
        """

        ...

    @cached_property
    def readable_id(self) -> str:
        """Human readable spectrometer id. Defaults to "<MODEL> - <SERIAL>" if
        not set by the implementing class.

        Returns
        -------
        str
        """
        return f"{self.model} - {self.serial_number}"

    def measure(self, repetitions: int = 1) -> ColorimeterMeasurement:
        """Trigger and collect one measurement from the LED tile

        Returns
        -------
        Measurement
            A dataclass containing various convenience calculations from python
            colour-science and the raw SPD.
        """

        if repetitions < 1:
            ArgumentError("Repetitions must be greater than 1")

        _rm: list[RawColorimeterMeasurement] = []
        for _ in range(repetitions):
            _rm += [self._raw_measure()]

        if len(_rm) == 1:
            return ColorimeterMeasurement.FromRaw(_rm[0])

        XYZ = np.asarray([m.XYZ for m in _rm]).mean(axis=0)
        exposure = np.mean([m.exposure for m in _rm]).item()
        id = _rm[0].device_id

        return ColorimeterMeasurement.FromRaw(
            RawColorimeterMeasurement(XYZ=XYZ, exposure=exposure, device_id=id)
        )


@final
class VirtualColorimeter(Colorimeter):
    """
    Basic spectroradiometer interface. Implements a virtual spectrometer
    returning random colors for basic testing
    """

    def __init__(self):
        super().__init__()

    @property
    def manufacturer(self) -> str:
        """Return "specio" as the manufacturer of this virtual spectrometer

        Returns
        -------
        str
        """
        return "specio"

    @cached_property
    def model(self):
        """The model name or model signature from the spectrometer.

        Returns
        -------
        str
        """
        return "Virtual Random Spectrometer"

    @property
    def serial_number(self):
        """The serial number of the spectrometer

        Returns
        -------
        str
        """
        return "0000-0000"

    def _raw_measure(self) -> RawColorimeterMeasurement:
        """Return a random SPD generated by combining three gaussian spectra

        Returns
        -------
        RawMeasurement
            A simple dataclass with the required parameters to produce fully
            defined :class:`specio.spectrometers.common.Measurement`
        """
        peaks = np.random.randint([460, 510, 600], [480, 570, 690], 3)
        widths = np.random.randint(40, 80, 3)
        powers = np.random.randint(10, 40, 3) / 1000
        spd = sd_multi_leds(
            peak_wavelengths=peaks,
            half_spectral_widths=widths / 2,
            peak_power_ratios=powers,
        )

        _measurement = RawColorimeterMeasurement(
            XYZ=sd_to_XYZ(spd, k=683),
            exposure=1.0,
            device_id="Virtual Spectrometer",
        )
        return _measurement



from pathlib import Path

import numpy as np
import pytest

from specio.common import VirtualSpectrometer
from specio.serialization.csmf import (
    CSMF_Data,
    CSMF_Metadata,
    csmf_data_to_buffer,
    load_csmf_file,
    save_csmf_file,
)


@pytest.fixture(scope="class")
def virtual_data() -> CSMF_Data:
    vspr = VirtualSpectrometer()

    NUM_VIRTUAL = 100
    measurements = [vspr.measure() for _ in range(NUM_VIRTUAL)]
    test_colors = np.random.uniform(0, 1023, (3, NUM_VIRTUAL)).astype(np.float32)
    order = np.random.permutation(NUM_VIRTUAL)

    ml = CSMF_Data(
        measurements=np.asarray(measurements),  # type: ignore
        test_colors=test_colors,
        order=order,
        metadata=CSMF_Metadata(
            notes="Random Test Measurements",
            author="tjdcs",
            location="virtual",
            software="specio-tests",
        ),
    )
    return ml


class Test_CSMF_Files:
    def test_csmf_rw(self, tmp_path: Path, virtual_data):
        p = tmp_path.joinpath("test_data")

        buffer = csmf_data_to_buffer(virtual_data)
        p = save_csmf_file(p, virtual_data)

        assert p.exists()
        assert p.is_file()
        assert p.read_bytes() == buffer.SerializeToString()

        read_data = load_csmf_file(p)

        assert read_data == virtual_data



"""
Implement support for operations with CR300
"""

import logging
import platform
import re
import textwrap
import time
from collections.abc import Mapping
from dataclasses import dataclass
from enum import Enum
from functools import cached_property
from types import MappingProxyType
from typing import Self, cast, final

import bidict
import numpy as np
import serial.tools.list_ports
from aenum import MultiValueEnum
from colour import SpectralDistribution, SpectralShape

from specio.common import RawSPDMeasurement, SpecRadiometer
from specio.common.colorimeters import Colorimeter, RawColorimeterMeasurement
from specio.common.utility import specio_warning

__author__ = "Tucker Downs"
__copyright__ = "Copyright 2022 Specio Developers"
__license__ = "MIT License - https://github.com/tjdcs/specio/blob/main/LICENSE.md"
__maintainer__ = "Tucker Downs"
__email__ = "tucker@tjdcs.dev"
__status__ = "Development"


_COMMAND_TIMEOUT = 0.05
_DEFAULT_SERIAL_TIMEOUT = 0.025
_CR_SERIAL_KWARGS: Mapping = MappingProxyType(
    {
        "baudrate": 115200,
        "bytesize": 8,
        "parity": "N",
        "rtscts": True,
        "timeout": _DEFAULT_SERIAL_TIMEOUT,
    }
)


class InstrumentType(MultiValueEnum):
    """
    Identifies the type of instrument
    =====
    0: Photometer
    1: Colorimeter
    2: Spectroradiometer
    """

    PHOTOMETER = 0, "0"
    COLORIMETER = 1, "1"
    SPECTRORADIOMETER = 2, "2"


class Model(Enum):
    """
    Identifies the CR model
    =====
    Values are based on the response to "RC Model" command.
    """

    CR300 = "CR-300"
    CR250 = "CR-250"


class ResponseType(bytes, Enum):
    """
    Identifies the success / failure of any CR command.
    """

    ERROR = b"ER"
    OK = b"OK"


class ResponseCode(int, MultiValueEnum):
    """
    Error codes for Colorimetry Research CR-300 and other model spectrometers
    =====
    """

    OK = 0
    INVALID_COMMAND = -500

    TOO_DARK = 100
    CANT_SYNC_CONST = 101
    CANT_SYNC_AUTO = 102
    SYNC_TOO_LOW = 103

    INVALID_SYNC_MODE = -300, -521
    INVALID_SYNC_PERIOD = -301
    CANT_SYNC_TO_LIGHT = -302

    LIGHT_INTENSITY_FLUCTUATION = -303
    LIGHT_INTENSITY_TOO_LOW = -304
    LIGHT_INTENSITY_UNMEASURABLE = -305
    LIGHT_INTENSITY_TOO_HIGH = -306

    HARDWARE_MALFUNCTION = -331
    MATRIX_VERSION_MISMATCH = -332
    INVALID_MATRIX_INDEX = -333

    NO_CIE_TABLES = -334
    NO_CMF_TABLES = -335

    NO_MATRIX_FOR_ID = -336

    DUPLICATE_FILTER_SELECTION = -505
    NO_ACCESSORY_FOR_INDEX = -506
    NO_FILTER_FOR_INDEX = -507
    INDEX_NOT_VALID_ACCESSORY = -508
    INDEX_NOT_VALID_FILTER = -509, -510, -511

    INVALID_RANGE_MODE = -512
    INVALID_RANGE_INDEX = -513
    INVALID_EXPOSURE_MULTIPLIER = -514

    INDEX_DOESNT_SELECT_APERTURE = -515

    INVALID_EXPOSURE_MODE = -518
    INVALID_EXPOSURE_VALUE = -519

    INVALID_SYNC_FREQUENCY = -522

    INVALID_MATRIX_MODE = -552
    INVALID_MATRIX_ID = -553
    INVALID_MATRIX_NAME = -555

    ERRROR_SAVING_MATRIX_FLASH = -559
    INVALID_USER_CALIBRATION_MODE = -560

    RESERVED = -999

    @classmethod
    def _missing_(cls, _: object):
        """
        Set the default error code to RESERVED
        =======================================
        """
        return cls.RESERVED


@dataclass
class CommandResponse:
    """
    Response data from Colorimetry Research
    """

    type: ResponseType
    code: ResponseCode
    description: str
    arguments: list[str]


class CommandError(Exception):
    """
    Describes an issue with sending a command to CR
    """

    def __init__(self, response: CommandResponse, *args: object) -> None:
        self.response = response
        super().__init__(*args)


@final
class CRSpectrometer(SpecRadiometer):
    """Interface with a colorimetry research brand CR-250 or CR-300. Implements
    the `specio.spectrometers.SpecRadiometer`

    Raises
    ------
    serial.SerialException
        if `CRSpectrometer.discovery` fails or there are other serial port issues.
    CommandError
        A error was encountered in parsing the result of the serial command to
        the hardware device.
    """

    class MeasurementSpeed(MultiValueEnum):
        """
        Controls the measurement speed when the CR Exposure Mode is set to "auto"
        """

        SLOW: Self = 0, "0", "slow"  # type: ignore
        NORMAL: Self = 1, "1", "normal"  # type: ignore
        FAST: Self = 2, "2", "fast"  # type: ignore
        FAST_2X: Self = 3, "3", "2x fast"  # type: ignore

    @classmethod
    def discover(cls) -> "CRSpectrometer":
        """Attempt automatic discovery of the CR serial port and return the
        CR spectrometer object.

        Returns
        -------
        CRSpectrometer
            A successfully automatic CS2000 object.

        Raises
        ------
        serial.SerialException
            If no serial port can be automatically linked.
        """
        if platform.system() == "Darwin":
            port_list = list(serial.tools.list_ports.grep("usbmodem"))
        elif platform.system() == "Windows":
            port_list = list(serial.tools.list_ports.grep("Colorimetry"))
        elif platform.system() == "Unix":
            raise NotImplementedError("CR discovery is not implemented for Unix")
        elif platform.system() == "Linux":
            port_list = list(serial.tools.list_ports.grep("ACM"))
        else:
            port_list = serial.tools.list_ports.comports()

        if len(port_list) == 0:
            raise serial.SerialException("No serial ports found on machine")

        for p in port_list:
            try:
                device = p.device  # type: ignore
                sp = serial.Serial(
                    device,
                    **{
                        "baudrate": 115200,
                        "bytesize": 8,
                        "parity": "N",
                        "rtscts": True,
                        "timeout": 0.1,
                    },
                )
                sp.read_all()
                sp.write(b"RC InstrumentType\n")

                response = sp.readline()
                if response.startswith(b"OK:0:RC InstrumentType:2"):
                    sp.close()
                    cr = CRSpectrometer(device)
                    return cr
            except:  # noqa: S112,E722
                continue

        raise serial.SerialException(
            textwrap.dedent(
                """Could not connect to any colorimetry research spectrometer.
                Check connection and device power."""
            )
        )

    def __init__(
        self,
        port: str,
        speed: MeasurementSpeed = MeasurementSpeed.NORMAL,
    ):
        """
        Construct CR Controller Obj
        """
        self.__last_cmd_time: float = 0
        if isinstance(port, str):
            self._port = serial.Serial(port, **_CR_SERIAL_KWARGS)

        self.measurement_speed = speed

    @property
    def manufacturer(self) -> str:
        """Return mfr name"""
        return "Colorimetry Research"

    @property
    def firmware(self) -> str:
        """The firmware version on the hardware

        Returns
        -------
        str
        """
        if not hasattr(self, "_firmware") or self._firmware is None:
            response = self._write_cmd("RC Firmware")
            self._firmware = response.arguments[0]
        return self._firmware

    @property
    def measurement_speed(self) -> MeasurementSpeed:
        """The automatic measurement speed of the hardware when in "auto" timing

        Returns
        -------
        MeasurementSpeed
        """
        response = self._write_cmd("SM ExposureMode 0")
        response = self._write_cmd("RS Speed")
        self._measurement_speed = CRSpectrometer.MeasurementSpeed(
            response.arguments[0].lower()
        )
        return self._measurement_speed

    @measurement_speed.setter
    def measurement_speed(self, speed: MeasurementSpeed):
        _ = self._write_cmd(f"SM Speed {speed.values[0]}")
        self._measurement_speed = speed

    @property
    def aperture(self):
        """
        Get spectrometer aperture value
        """
        if not hasattr(self, "_aperture") or self._aperture is None:
            response = self._write_cmd("RS Aperture")
            self._aperture = response.arguments[0]
        return self._aperture

    @property
    def serial_number(self) -> str:
        """The hardware serial number

        Returns
        -------
        str
        """
        if not hasattr(self, "_sn") or self._sn is None:
            response = self._write_cmd("RC ID")
            self._sn = response.arguments[0]
        return self._sn

    @property
    def average_samples(self) -> int:
        response = self._write_cmd("RS ExposureX")
        return int(response.arguments[0])

    @average_samples.setter
    def average_samples(self, num: int) -> None:
        num = num if num > 0 else 1
        num = num if num < 50 else 50
        self._write_cmd(f"SM ExposureX {num:d}")

    @cached_property
    def model(self) -> str:
        """The model name

        Returns
        -------
        str
        """
        response = self._write_cmd("RC Model")
        return response.arguments[0]

    @property
    def instrument_type(self):
        """
        Check that the connected device is a spectrometer
        """
        if not hasattr(self, "_instrument_type") or self._instrument_type is None:
            response = self._write_cmd("RC InstrumentType")
            i_type = InstrumentType(response.arguments[0])
            self._instrument_type = i_type

        return self._instrument_type

    def __clear_buffer(self):
        """
        Clear input buffer
        """
        t = self._port.timeout
        self._port.apply_settings({"timeout": _DEFAULT_SERIAL_TIMEOUT})
        self._port.readall()
        self._port.apply_settings({"timeout": t})

    def _write_cmd(self, command: str) -> CommandResponse:
        """
        Write cmd to serial port
        """
        log = logging.getLogger("specio.CR")
        log.debug("Sending CMD: %s", command)

        enc_command: bytes = (command + "\n").encode()
        if self.__last_cmd_time + _COMMAND_TIMEOUT > time.time():
            time.sleep(
                max(
                    self.__last_cmd_time + _COMMAND_TIMEOUT + 0.001 - time.time(),
                    0,
                )
            )

        self.__clear_buffer()
        self._port.write(enc_command)
        self.__last_cmd_time = time.time()

        response = self._port.readline()

        response = self._parse_response(response)

        if response.type == ResponseType.ERROR:
            raise CommandError(response, response.arguments[0])
        else:
            return response

    def _parse_response(self, data: bytes) -> CommandResponse:
        """
        Parse CR response string
        """
        response = data.strip().split(b":")

        args = []
        if (
            response[3].decode().isnumeric()
            and int(response[3]) > 0
            and self._port.in_waiting
        ):
            for _ in range(int(response[3])):
                n_response = self._port.readline()
                args.append(n_response)
        else:
            args = [r.decode() for r in response[3:]]

        return CommandResponse(
            type=ResponseType(response[0]),
            code=ResponseCode(int(response[1].decode())),
            description=response[2].decode(),
            arguments=args,
        )

    def _apply_measurementspeed_timeout(self):
        if self.measurement_speed is CRSpectrometer.MeasurementSpeed.SLOW:
            t = 70
        elif self.measurement_speed is CRSpectrometer.MeasurementSpeed.NORMAL:
            t = 21
        elif self.measurement_speed is CRSpectrometer.MeasurementSpeed.FAST:
            t = 14
        else:
            t = 7
        t *= self.average_samples
        self._port.apply_settings({"timeout": t})

    def _raw_measure(self) -> RawSPDMeasurement:
        """
        Make spectral measurement with CR
        """
        t = self._port.timeout

        self._apply_measurementspeed_timeout()
        response = self._write_cmd("M")
        self._port.apply_settings({"timeout": t})

        self._port.apply_settings({"timeout": 0.31})
        response = self._write_cmd("RM Spectrum")
        self._port.apply_settings({"timeout": t})

        args = response.arguments[0].split(",")
        if float(args[1]) != 0:
            shape = SpectralShape(
                start=float(args[0]),
                end=float(args[1]),
                interval=float(args[2]),
            )
        elif self.model == "CR-300":
            shape = SpectralShape(380, 780, 1)
        elif self.model == "CR-250":
            shape = SpectralShape(380, 780, 4)

        time.sleep(0.01)
        data = [self._port.readline() for _ in range(len(shape.wavelengths))]
        data = [float(d.decode()) for d in data]

        exposure = self._write_cmd("RM Exposure").arguments[0]
        exMatch = re.match(r"\d*\.?\d*", exposure)
        exposure = float(exMatch.group()) / 1000 if exMatch else -1

        return RawSPDMeasurement(
            spd=SpectralDistribution(data=data, domain=shape),
            spectrometer_id=self.readable_id,
            exposure=exposure,
        )


@final
class CRColorimeter(Colorimeter):
    """Interface with a colorimetry research brand CR-250 or CR-300. Implements
    the `specio.spectrometers.SpecRadiometer`

    Raises
    ------
    serial.SerialException
        if `CRSpectrometer.discovery` fails or there are other serial port issues.
    CommandError
        A error was encountered in parsing the result of the serial command to
        the hardware device.
    """

    @classmethod
    def discover(cls) -> "CRColorimeter":
        """Attempt automatic discovery of the CR serial port and return the
        CR spectrometer object.

        Returns
        -------
        CRSpectrometer
            A successfully automatic CS2000 object.

        Raises
        ------
        serial.SerialException
            If no serial port can be automatically linked.
        """
        if platform.system() == "Darwin":
            port_list = list(serial.tools.list_ports.grep("usbmodem"))
        elif platform.system() == "Windows":
            port_list = list(serial.tools.list_ports.grep("Colorimetry"))
        elif platform.system() == "Unix":
            raise NotImplementedError("CR discovery is not implemented for Unix")
        elif platform.system() == "Linux":
            port_list = list(serial.tools.list_ports.grep("ACM"))
        else:
            port_list = serial.tools.list_ports.comports()

        if len(port_list) == 0:
            raise serial.SerialException("No serial ports found on machine")

        for p in port_list:
            try:
                device = p.device  # type: ignore
                sp = serial.Serial(device, **_CR_SERIAL_KWARGS)
                sp.readall()
                sp.write(b"RC InstrumentType\n")

                response = sp.readline()
                if response.startswith(b"OK:0:RC InstrumentType:1"):
                    sp.close()
                    cr = CRColorimeter(device)
                    return cr
            except:  # noqa: S112,E722
                continue

        raise serial.SerialException(
            textwrap.dedent(
                """Could not connect to any colorimetry research colorimeter.
                Check connection and device power."""
            )
        )

    def __init__(
        self,
        port: str,
    ):
        """
        Construct CR Controller Obj
        """
        self.__last_cmd_time: float = 0
        if isinstance(port, str):
            self._port = serial.Serial(port, **_CR_SERIAL_KWARGS)

        self._warn_filter_selection()

    @property
    def manufacturer(self) -> str:
        """Return mfr name"""
        return "Colorimetry Research"

    @property
    def firmware(self) -> str:
        """The firmware version on the hardware

        Returns
        -------
        str
        """
        if not hasattr(self, "_firmware") or self._firmware is None:
            response = self._write_cmd("RC Firmware")
            self._firmware = response.arguments[0]
        return self._firmware

    @property
    def aperture(self):
        """
        Get spectrometer aperture value
        """
        if not hasattr(self, "_aperture") or self._aperture is None:
            response = self._write_cmd("RS Aperture")
            self._aperture = response.arguments[0]
        return self._aperture

    @property
    def serial_number(self) -> str:
        """The hardware serial number

        Returns
        -------
        str
        """
        if not hasattr(self, "_sn") or self._sn is None:
            response = self._write_cmd("RC ID")
            self._sn = response.arguments[0]
        return self._sn

    @cached_property
    def available_filters(self) -> bidict.bidict[int, str]:
        response = self._write_cmd("RC Filter")
        filters = bidict.bidict()
        for arg in response.arguments:
            arg = cast(bytes, arg)
            items = arg.decode().strip().split(",")
            filters[int(items[0])] = items[1]
        filters[0] = "None"
        return filters

    @property
    def current_filters(self) -> tuple[int, int, int]:
        response = self._write_cmd("RS Filter")
        arguments = response.arguments[0].split(",")

        out = []
        for f in arguments:
            out += [self.available_filters.inverse[f]]
        return tuple(out)

    @current_filters.setter
    def current_filters(self, filters: tuple[int, ...]):
        if len(filters) > 3:
            RuntimeError("CR-100/120 only supports up to 3 filter selectons!")
        for i in range(1, 4):
            cur_filter_id = filters[i - 1] if i <= len(filters) else -1
            self._write_cmd(f"SM Filter{i:.0f} {cur_filter_id:.0f}")

        self._warn_filter_selection()

    @property
    def current_filters_names(self) -> tuple[str, str, str]:
        # ignore type error. Cannot interpret list builder size.
        return tuple([self.available_filters[k] for k in self.current_filters])  # type: ignore

    def _warn_filter_selection(self):
        cur = self.current_filters
        if len(cur) == 0:
            specio_warning("Check colorimeter has no active filters.")
        elif len(cur) == 1:
            specio_warning(
                f"Check colorimeter has one filter: {self.available_filters[cur[0]]}"
            )
        else:
            filters_string = ", ".join([self.available_filters[f] for f in cur])
            specio_warning(f"Check colorimeter has stacked filters: {filters_string}.")

    @cached_property
    def model(self) -> str:
        """The model name

        Returns
        -------
        str
        """
        response = self._write_cmd("RC Model")
        return response.arguments[0]

    @property
    def average_samples(self) -> int:
        """
        Check that the connected device is a spectrometer
        """

        response = self._write_cmd("RS ExposureX")
        return int(response.arguments[0])

    @average_samples.setter
    def average_samples(self, num: int):
        num = num if num > 0 else 1
        num = num if num < 50 else 50
        self._write_cmd(f"SM ExposureX {num:d}")

    @property
    def instrument_type(self):
        """
        Check that the connected device is a spectrometer
        """
        if not hasattr(self, "_instrument_type") or self._instrument_type is None:
            response = self._write_cmd("RC InstrumentType")
            i_type = InstrumentType(response.arguments[0])
            self._instrument_type = i_type

        return self._instrument_type

    def __clear_buffer(self):
        """
        Clear input buffer
        """
        t = self._port.timeout
        self._port.apply_settings({"timeout": _DEFAULT_SERIAL_TIMEOUT})
        self._port.readall()
        self._port.apply_settings({"timeout": t})

    def _write_cmd(self, command: str) -> CommandResponse:
        """
        Write cmd to serial port
        """
        log = logging.getLogger("specio.CR")
        log.debug("Sending CMD: %s", command)

        enc_command: bytes = (command + "\n").encode()
        if self.__last_cmd_time + _COMMAND_TIMEOUT > time.time():
            time.sleep(
                max(
                    self.__last_cmd_time + _COMMAND_TIMEOUT + 0.001 - time.time(),
                    0,
                )
            )

        self.__clear_buffer()
        self._port.write(enc_command)
        self.__last_cmd_time = time.time()

        response = self._port.readline()

        response = self._parse_response(response)

        if response.type == ResponseType.ERROR:
            raise CommandError(response, response.arguments[0])
        else:
            return response

    def _parse_response(self, data: bytes) -> CommandResponse:
        """
        Parse CR response string
        """
        response = data.strip().split(b":")

        args = []
        if (
            response[3].decode().isnumeric()
            and int(response[3]) > 0
            and self._port.in_waiting
        ):
            for _ in range(int(response[3])):
                n_response = self._port.readline()
                args.append(n_response)
        else:
            args = [r.decode() for r in response[3:]]

        return CommandResponse(
            type=ResponseType(response[0]),
            code=ResponseCode(int(response[1].decode())),
            description=response[2].decode(),
            arguments=args,
        )

    def _raw_measure(self) -> RawColorimeterMeasurement:
        """
        Make spectral measurement with CR
        """
        t = self._port.timeout

        self._port.apply_settings({"timeout": 10 + 0.5 * self.average_samples})
        response = self._write_cmd("M")

        self._port.apply_settings({"timeout": 0.21})
        response = self._write_cmd("RM XYZ")

        XYZ = np.asarray([float(s) for s in response.arguments[0].split(",")])

        exposure = self._write_cmd("RM Exposure").arguments[0]
        exMatch = re.match(r"\d*\.?\d*", exposure)
        exposure = float(exMatch.group()) / 1000 if exMatch else -1

        self._port.apply_settings({"timeout": t})

        return RawColorimeterMeasurement(
            XYZ=XYZ,
            device_id=self.readable_id,
            exposure=exposure,
        )



from specio.serialization.protobuf import common_pb2 as _common_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Colorimeter_Measurement(_message.Message):
    __slots__ = ("XYZ", "xy", "cct", "dominant_wl", "purity", "time", "exposure", "colorimeter_id")
    XYZ_FIELD_NUMBER: _ClassVar[int]
    XY_FIELD_NUMBER: _ClassVar[int]
    CCT_FIELD_NUMBER: _ClassVar[int]
    DOMINANT_WL_FIELD_NUMBER: _ClassVar[int]
    PURITY_FIELD_NUMBER: _ClassVar[int]
    TIME_FIELD_NUMBER: _ClassVar[int]
    EXPOSURE_FIELD_NUMBER: _ClassVar[int]
    COLORIMETER_ID_FIELD_NUMBER: _ClassVar[int]
    XYZ: _common_pb2.XYZ_value
    xy: _common_pb2.xy_value
    cct: _common_pb2.cct_value
    dominant_wl: float
    purity: float
    time: _common_pb2.Timestamp
    exposure: float
    colorimeter_id: str
    def __init__(self, XYZ: _Optional[_Union[_common_pb2.XYZ_value, _Mapping]] = ..., xy: _Optional[_Union[_common_pb2.xy_value, _Mapping]] = ..., cct: _Optional[_Union[_common_pb2.cct_value, _Mapping]] = ..., dominant_wl: _Optional[float] = ..., purity: _Optional[float] = ..., time: _Optional[_Union[_common_pb2.Timestamp, _Mapping]] = ..., exposure: _Optional[float] = ..., colorimeter_id: _Optional[str] = ...) -> None: ...

class SPD_Measurement(_message.Message):
    __slots__ = ("spd", "exposure", "XYZ", "xy", "cct", "dominant_wl", "purity", "power", "spectrometer_id", "time")
    SPD_FIELD_NUMBER: _ClassVar[int]
    EXPOSURE_FIELD_NUMBER: _ClassVar[int]
    XYZ_FIELD_NUMBER: _ClassVar[int]
    XY_FIELD_NUMBER: _ClassVar[int]
    CCT_FIELD_NUMBER: _ClassVar[int]
    DOMINANT_WL_FIELD_NUMBER: _ClassVar[int]
    PURITY_FIELD_NUMBER: _ClassVar[int]
    POWER_FIELD_NUMBER: _ClassVar[int]
    SPECTROMETER_ID_FIELD_NUMBER: _ClassVar[int]
    TIME_FIELD_NUMBER: _ClassVar[int]
    spd: _common_pb2.SpectralDistribution
    exposure: float
    XYZ: _common_pb2.XYZ_value
    xy: _common_pb2.xy_value
    cct: _common_pb2.cct_value
    dominant_wl: float
    purity: float
    power: float
    spectrometer_id: str
    time: _common_pb2.Timestamp
    def __init__(self, spd: _Optional[_Union[_common_pb2.SpectralDistribution, _Mapping]] = ..., exposure: _Optional[float] = ..., XYZ: _Optional[_Union[_common_pb2.XYZ_value, _Mapping]] = ..., xy: _Optional[_Union[_common_pb2.xy_value, _Mapping]] = ..., cct: _Optional[_Union[_common_pb2.cct_value, _Mapping]] = ..., dominant_wl: _Optional[float] = ..., purity: _Optional[float] = ..., power: _Optional[float] = ..., spectrometer_id: _Optional[str] = ..., time: _Optional[_Union[_common_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class CSFM_File(_message.Message):
    __slots__ = ("measurements", "spd_measurements", "test_colors", "order", "notes", "author", "location", "software", "ancillary")
    class Measurement(_message.Message):
        __slots__ = ("spd", "xyz")
        SPD_FIELD_NUMBER: _ClassVar[int]
        XYZ_FIELD_NUMBER: _ClassVar[int]
        spd: SPD_Measurement
        xyz: Colorimeter_Measurement
        def __init__(self, spd: _Optional[_Union[SPD_Measurement, _Mapping]] = ..., xyz: _Optional[_Union[Colorimeter_Measurement, _Mapping]] = ...) -> None: ...
    class TestColor(_message.Message):
        __slots__ = ("c", "f")
        C_FIELD_NUMBER: _ClassVar[int]
        F_FIELD_NUMBER: _ClassVar[int]
        c: _containers.RepeatedScalarFieldContainer[int]
        f: _containers.RepeatedScalarFieldContainer[float]
        def __init__(self, c: _Optional[_Iterable[int]] = ..., f: _Optional[_Iterable[float]] = ...) -> None: ...
    MEASUREMENTS_FIELD_NUMBER: _ClassVar[int]
    SPD_MEASUREMENTS_FIELD_NUMBER: _ClassVar[int]
    TEST_COLORS_FIELD_NUMBER: _ClassVar[int]
    ORDER_FIELD_NUMBER: _ClassVar[int]
    NOTES_FIELD_NUMBER: _ClassVar[int]
    AUTHOR_FIELD_NUMBER: _ClassVar[int]
    LOCATION_FIELD_NUMBER: _ClassVar[int]
    SOFTWARE_FIELD_NUMBER: _ClassVar[int]
    ANCILLARY_FIELD_NUMBER: _ClassVar[int]
    measurements: _containers.RepeatedCompositeFieldContainer[CSFM_File.Measurement]
    spd_measurements: _containers.RepeatedCompositeFieldContainer[SPD_Measurement]
    test_colors: _containers.RepeatedCompositeFieldContainer[CSFM_File.TestColor]
    order: _containers.RepeatedScalarFieldContainer[int]
    notes: str
    author: str
    location: str
    software: str
    ancillary: bytes
    def __init__(self, measurements: _Optional[_Iterable[_Union[CSFM_File.Measurement, _Mapping]]] = ..., spd_measurements: _Optional[_Iterable[_Union[SPD_Measurement, _Mapping]]] = ..., test_colors: _Optional[_Iterable[_Union[CSFM_File.TestColor, _Mapping]]] = ..., order: _Optional[_Iterable[int]] = ..., notes: _Optional[str] = ..., author: _Optional[str] = ..., location: _Optional[str] = ..., software: _Optional[str] = ..., ancillary: _Optional[bytes] = ...) -> None: ...



"""
Provides serialization for Measurement Classes
"""

import datetime
from typing import cast

import numpy as np
from colour import SpectralDistribution, SpectralShape

from specio.common import ColorimeterMeasurement, SPDMeasurement
from specio.serialization.protobuf import common_pb2, measurements_pb2


def colorimeter_measurement_to_proto(
    col: ColorimeterMeasurement,
) -> measurements_pb2.Colorimeter_Measurement:
    buf = measurements_pb2.Colorimeter_Measurement(
        XYZ=common_pb2.XYZ_value(X=col.XYZ[0], Y=col.XYZ[1], Z=col.XYZ[2]),
        xy=common_pb2.xy_value(x=col.xy[0], y=col.xy[1]),
        cct=common_pb2.cct_value(cct=col.cct, duv=col.duv),
        dominant_wl=col.dominant_wl,
        purity=col.purity,
        time=common_pb2.Timestamp(timestr=col.time.isoformat()),
        exposure=col.exposure,
        colorimeter_id=col.device_id,
    )
    return buf


def colorimeter_measurement_to_bytes(col: ColorimeterMeasurement) -> bytes:
    return colorimeter_measurement_to_proto(col).SerializeToString()


def colorimeter_measurement_from_bytes(
    buffer: bytes | measurements_pb2.Colorimeter_Measurement,
    recompute: bool = False,
) -> ColorimeterMeasurement:
    if isinstance(buffer, bytes):
        buffer = measurements_pb2.Colorimeter_Measurement.FromString(buffer)
    buffer = cast(measurements_pb2.Colorimeter_Measurement, buffer)
    cm = ColorimeterMeasurement(
        XYZ=np.asarray((buffer.XYZ.X, buffer.XYZ.Y, buffer.XYZ.Z)),
        exposure=buffer.exposure,
        device_id=buffer.colorimeter_id,
        no_compute=not recompute,
    )

    if not recompute:
        cm.cct = buffer.cct.cct
        cm.duv = buffer.cct.duv
        cm.xy = np.asarray((buffer.xy.x, buffer.xy.y))
        cm.dominant_wl = buffer.dominant_wl
        cm.purity = buffer.purity
    cm.time = datetime.datetime.fromisoformat(buffer.time.timestr)
    return cm


def spd_measurement_to_proto(
    spd: SPDMeasurement,
) -> measurements_pb2.SPD_Measurement:
    """
    Transform input :class:SPDMeasurement to a protobuf. Proto defined by
    specio/serialization/protobuf/measurements.proto

    Parameters
    ----------
    spd : SPDMeasurement
        A complete measurement object

    Returns
    -------
    measurements_pb2.SPD_Measurement
        proto handle for serializing the SPD Data.
    """
    shape_buf = common_pb2.SpectralShape(
        start=spd.spd.shape.start,
        end=spd.spd.shape.end,
        step=spd.spd.shape.interval,
    )
    spd_buf = common_pb2.SpectralDistribution(
        shape=shape_buf, values=spd.spd.values, name=spd.spd.name
    )

    buf = measurements_pb2.SPD_Measurement(
        spd=spd_buf,
        exposure=spd.exposure,
        XYZ=common_pb2.XYZ_value(X=spd.XYZ[0], Y=spd.XYZ[1], Z=spd.XYZ[2]),
        xy=common_pb2.xy_value(x=spd.xy[0], y=spd.xy[1]),
        cct=common_pb2.cct_value(cct=spd.cct, duv=spd.duv),
        dominant_wl=spd.dominant_wl,
        purity=spd.purity,
        power=spd.power,
        spectrometer_id=spd.spectrometer_id,
        time=common_pb2.Timestamp(timestr=spd.time.isoformat()),
    )
    return buf


def spd_measurement_to_bytes(spd: SPDMeasurement) -> bytes:
    """Transform input :class:SPDMeasurement to a bytes string. Returned byte
    string is readable as a protobuf defined by
    specio/serialization/protobuf/measurements.proto

    Parameters
    ----------
    spd : SPDMeasurement
        A complete measurement object

    Returns
    -------
    bytes
        a string of bytes serialized via measurements.proto
    """

    return spd_measurement_to_proto(spd).SerializeToString()


def spd_measurement_from_bytes(
    buffer: bytes | measurements_pb2.SPD_Measurement, recompute: bool = False
) -> SPDMeasurement:
    """Transform a bytes string or protobuffer handle to a SPDMeasurements
    object.

    Parameters
    ----------
    buffer : bytes | measurements_pb2.SPD_Measurement
        data string or protobuffer handle object.
    recompute : bool, optional
        Optionally recompute derivable data from the spectral data. Does not read
        XYZ, CCT, etc... from the data string. Default False

    Returns
    -------
    SPDMeasurement
        The measurement data de-serialized.

    See Also
    --------
    spd_measurement_to_bytes
    """
    if isinstance(buffer, bytes):
        buffer = measurements_pb2.SPD_Measurement.FromString(buffer)
    buffer = cast(measurements_pb2.SPD_Measurement, buffer)

    spd = SpectralDistribution(
        (buffer.spd.values if len(buffer.spd.values) > 0 else buffer.spd.values_old),
        domain=SpectralShape(
            buffer.spd.shape.start, buffer.spd.shape.end, buffer.spd.shape.step
        ),
    )

    ret = SPDMeasurement(
        spd, buffer.exposure, buffer.spectrometer_id, no_compute=not recompute
    )
    if not recompute:
        ret.cct = buffer.cct.cct
        ret.duv = buffer.cct.duv
        ret.XYZ = np.asarray((buffer.XYZ.X, buffer.XYZ.Y, buffer.XYZ.Z))
        ret.xy = np.asarray((buffer.xy.x, buffer.xy.y))
        ret.dominant_wl = buffer.dominant_wl
        ret.purity = buffer.purity
        ret.power = buffer.power
    ret.time = datetime.datetime.fromisoformat(buffer.time.timestr)
    return ret



from typing import cast

import numpy as np
from colour import SpectralDistribution, SpectralShape

from specio.serialization.protobuf import common_pb2

__all__ = [
    "sd_to_buffer",
    "buffer_to_sd",
    "sd_shape_to_buffer",
    "buffer_to_sd_shape",
]


def sd_shape_to_buffer(shape: SpectralShape) -> common_pb2.SpectralShape:
    """Convert SpectralShape to buffer. Defaults to bytes but may optionally
    return protobuf object

    Parameters
    ----------
    shape : SpectralShape
        The spectral shape to convert
    return_pb : bool, optional
        Change return type, by default False

    Returns
    -------
    bytes | protobuf.SpectralShape
        SpectralShape, serialized to `bytes` unless `return_pb` is True

    See Also
    --------
    `buffer_to_sd_shape`
    """
    pb = common_pb2.SpectralShape()
    pb.start = shape.start
    pb.end = shape.end
    pb.step = shape.interval

    return pb


def buffer_to_sd_shape(
    buffer: bytes | common_pb2.SpectralShape,
) -> SpectralShape:
    """
    Convert buffer back to `SpectralDistribution`

    Parameters
    ----------
    buffer : bytes | protobuf.SpectralShape
        the buffer to convert. Buffer must be serialized by sd_shape_to_buffer
        or compatible external program.

    Returns
    -------
    SpectralShape

    See Also
    --------
    `sd_shape_to_buffer`
    """

    if isinstance(buffer, bytes):
        buffer = common_pb2.SpectralShape.FromString(buffer)
    buffer = cast(common_pb2.SpectralShape, buffer)

    return SpectralShape(start=buffer.start, end=buffer.end, interval=buffer.step)


def sd_to_buffer(
    sd: SpectralDistribution, return_pb: bool = False
) -> bytes | common_pb2.SpectralDistribution:
    """Convert a `SpectralDistribution` object to a buffer.

    Parameters
    ----------
    sd : SpectralDistribution
        The spectral distribution to convert
    return_pb: bool = False
        changes return type to python protobuf object

    Returns
    -------
    bytes
        a byte or proto buffer that can be decoded back into a SpectralDistribution
        object via `specio.io.sd_from_buffer`
    """

    pb = common_pb2.SpectralDistribution()
    pb.name = sd.name

    pb.shape = sd_shape_to_buffer(sd.shape)

    pb.values.extend(sd.values.astype(np.double).tolist())
    pb.name = sd.name

    return pb if return_pb else pb.SerializeToString()


def buffer_to_sd(
    buffer: bytes | common_pb2.SpectralDistribution,
) -> SpectralDistribution:
    """Try to convert a byte buffer to SpectralDistribution

    Parameters
    ----------
    buffer : bytes
        The buffer to convert. Must come from `sd_to_buffer`

    Returns
    -------
    SpectralDistribution
        The spectral distribution

    Raises
    ------
    NotImplementedError
        _description_
    """

    if isinstance(buffer, bytes):
        pb = common_pb2.SpectralDistribution()
        pb.ParseFromString(buffer)
        buffer = pb
    pb = cast(common_pb2.SpectralDistribution, buffer)

    shape = buffer_to_sd_shape(pb.shape)
    values = pb.values if len(pb.values) > 0 else pb.values_old

    return SpectralDistribution(
        data=values,
        domain=shape,
        name=pb.name,
    )



from specio.common import VirtualColorimeter, VirtualSpectrometer
from specio.serialization.measurements import (
    colorimeter_measurement_from_bytes,
    colorimeter_measurement_to_bytes,
    spd_measurement_from_bytes,
    spd_measurement_to_bytes,
)


class TestSPDMeasurementSerialization:
    def test_spd_to_bytes(self):
        vs = VirtualSpectrometer()
        m = vs.measure()
        result = spd_measurement_to_bytes(m)

        assert isinstance(result, bytes)
        assert len(result) > 0

    def test_round_trip(self):
        vs = VirtualSpectrometer()
        m = vs.measure()

        result = spd_measurement_to_bytes(m)

        m2 = spd_measurement_from_bytes(result)
        assert m2 == m


class TestColorimeterSerialization:
    def test_col_measurement_to_bytes(self):
        vc = VirtualColorimeter()
        m = vc.measure()

        data = colorimeter_measurement_to_bytes(m)

        assert isinstance(data, bytes)
        assert len(data) > 0

    def test_round_trip(self):
        vc = VirtualColorimeter()
        m = vc.measure()

        data = colorimeter_measurement_to_bytes(m)

        assert isinstance(data, bytes)

        m2 = colorimeter_measurement_from_bytes(data)

        assert m == m2



#!python3
"""
CSMF Doctor fixes some broken CSMF files by recalculating derived attributes.
"""
import argparse
import textwrap
from pathlib import Path

from specio.serialization import csmf


def main():
    parser = argparse.ArgumentParser(
        prog="CSMF Doctor",
        description=textwrap.dedent(
            """Read CSMF files and recompute derived data. Output to the same files.
            If a directory is given, all CSMF files will be procesed.
            """
        ),
    )

    parser.add_argument(
        "-r",
        action="store_true",
        help="""Recursive. If file is a directory, all subdirectories will be
            searched searched.""",
    )

    parser.add_argument(
        "file",
        default=".",
        help="""The file to recompute. If file is a directory, all CSMF files will
            be updated. See also -r""",
    )

    args = parser.parse_args()

    base = Path(args.file)

    if not base.exists():
        raise RuntimeError(f"File or Directory {base!s} does not exist.")

    if base.is_dir() and args.r:
        files = tuple(base.glob("**/*.csmf"))
    elif base.is_dir():
        files = tuple(base.glob("*.csmf"))
    else:
        files = (base,)

    if len(files) == 0:
        print("No csmf files found.")

    for file in files:
        print(f"Processing Next File: {file.name!s}")

        data = csmf.load_csmf_file(file, recompute=True)
        csmf.save_csmf_file(file, data)

    print("All files processed")


if __name__ == "__main__":
    main()



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SpectralShape(_message.Message):
    __slots__ = ("start", "end", "step")
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    STEP_FIELD_NUMBER: _ClassVar[int]
    start: float
    end: float
    step: float
    def __init__(self, start: _Optional[float] = ..., end: _Optional[float] = ..., step: _Optional[float] = ...) -> None: ...

class SpectralDistribution(_message.Message):
    __slots__ = ("shape", "values", "values_old", "name")
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    VALUES_OLD_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    shape: SpectralShape
    values: _containers.RepeatedScalarFieldContainer[float]
    values_old: _containers.RepeatedScalarFieldContainer[float]
    name: str
    def __init__(self, shape: _Optional[_Union[SpectralShape, _Mapping]] = ..., values: _Optional[_Iterable[float]] = ..., values_old: _Optional[_Iterable[float]] = ..., name: _Optional[str] = ...) -> None: ...

class XYZ_value(_message.Message):
    __slots__ = ("X", "Y", "Z")
    X_FIELD_NUMBER: _ClassVar[int]
    Y_FIELD_NUMBER: _ClassVar[int]
    Z_FIELD_NUMBER: _ClassVar[int]
    X: float
    Y: float
    Z: float
    def __init__(self, X: _Optional[float] = ..., Y: _Optional[float] = ..., Z: _Optional[float] = ...) -> None: ...

class xy_value(_message.Message):
    __slots__ = ("x", "y")
    X_FIELD_NUMBER: _ClassVar[int]
    Y_FIELD_NUMBER: _ClassVar[int]
    x: float
    y: float
    def __init__(self, x: _Optional[float] = ..., y: _Optional[float] = ...) -> None: ...

class cct_value(_message.Message):
    __slots__ = ("cct", "duv")
    CCT_FIELD_NUMBER: _ClassVar[int]
    DUV_FIELD_NUMBER: _ClassVar[int]
    cct: float
    duv: float
    def __init__(self, cct: _Optional[float] = ..., duv: _Optional[float] = ...) -> None: ...

class Timestamp(_message.Message):
    __slots__ = ("timestr",)
    TIMESTR_FIELD_NUMBER: _ClassVar[int]
    timestr: str
    def __init__(self, timestr: _Optional[str] = ...) -> None: ...



from .colorimeters import (
    Colorimeter,
    ColorimeterMeasurement,
    RawColorimeterMeasurement,
    VirtualColorimeter,
)
from .spectrometers import (
    RawSPDMeasurement,
    SPDMeasurement,
    SpecRadiometer,
    VirtualSpectrometer,
)
from .utility import SuspiciousFileOperationError, get_valid_filename

__version__ = "0.2.9"
__author__ = "Tucker Downs"
__copyright__ = "Copyright 2022 Specio Developers"
__license__ = "MIT License - https://github.com/tjdcs/specio/blob/main/LICENSE.md"
__maintainer__ = "Tucker Downs"
__email__ = "tucker@tjdcs.dev"
__status__ = "Development"



import re
from typing import Any

from colour.utilities import warning


class SuspiciousFileOperationError(Exception):
    """Generated when a user does something suspicious with file names"""


def get_valid_filename(name: str) -> str:
    """Clean / validate filename string

    Parameters
    ----------
    name : str
        The string to be cleaned for file name validity

    Returns
    -------
    str
        A clean filename

    Raises
    ------
    SuspiciousFileOperation
        if the cleaned string looks like a spooky filepath (i.e. '/', '.', etc...)
    """
    s = str(name).strip().replace(" ", "_")
    s = re.sub(r"(?u)[^-\w.]", "", s)
    s = re.sub(r"_+-+_+", "__", s)
    if s in {"", ".", ".."}:
        raise SuspiciousFileOperationError(f"Could not derive file name from '{name}'")
    return s


class SpecioRuntimeWarning(Warning):
    """
    Define the base class of *Colour* runtime warnings.

    It is a subclass of the :class:`colour.utilities.ColourWarning` class.
    """


def specio_warning(*args: Any, **kwargs: Any):
    """
    Issue a specio runtime warning.

    Other Parameters
    ----------------
    args
        Arguments.
    kwargs
        Keywords arguments.

    Examples
    --------
    >>> usage_warning("This is a runtime warning!")  # doctest: +SKIP
    """

    kwargs["category"] = SpecioRuntimeWarning

    warning(*args, **kwargs)



"""
Define basic spectrometer interfaces
"""

import textwrap
from abc import ABC, abstractmethod
from ctypes import ArgumentError
from dataclasses import dataclass
from datetime import datetime
from functools import cached_property
from typing import Any, Self, final

import numpy as np
from colour import SpectralDistribution, SpragueInterpolator, sd_multi_leds
from colour.colorimetry.dominant import (
    colorimetric_purity,
    dominant_wavelength,
)
from colour.colorimetry.tristimulus_values import sd_to_XYZ
from colour.models.cie_ucs import xy_to_UCS_uv
from colour.models.cie_xyy import XYZ_to_xy
from colour.temperature import uv_to_CCT

__author__ = "Tucker Downs"
__copyright__ = "Copyright 2022 Specio Developers"
__license__ = "MIT License - https://github.com/tjdcs/specio/blob/main/LICENSE.md"
__maintainer__ = "Tucker Downs"
__email__ = "tucker@tjdcs.dev"
__status__ = "Development"

__all__ = []


@dataclass
class RawSPDMeasurement:
    """
    A measurement only containing the bare minimum amount of collected data.
    Expected to be further transformed into `specio.spectrometers.common.Measurement`.
    """

    spd: SpectralDistribution
    exposure: float
    spectrometer_id: str
    anc_data: Any = None


class SPDMeasurement:
    """
    The basic measurement structure for specio. It contains
    `colour.SpectralDistribution` and several convenience calculations. It an
    also be converted to a readable string for logging.
    """

    @classmethod
    def FromRaw(cls, raw: RawSPDMeasurement) -> Self:
        return cls(
            spd=raw.spd,
            exposure=raw.exposure,
            spectrometer_id=raw.spectrometer_id,
            ancillary=raw.anc_data,
        )

    def __init__(
        self,
        spd: SpectralDistribution,
        exposure: float,
        spectrometer_id: str,
        ancillary: Any = None,
        no_compute: bool = False,
    ):
        self.spd = spd
        self.exposure = exposure
        self.spectrometer_id = spectrometer_id
        self.anc_data = ancillary

        method = "ASTM E308"
        if self.spd.shape.interval not in [1, 5, 10, 20]:
            method = "Integration"

        if not no_compute:
            self.XYZ = sd_to_XYZ(self.spd, k=683, method=method)
            self.xy = XYZ_to_xy(self.XYZ)
            _cct = uv_to_CCT(xy_to_UCS_uv(self.xy))
            self.cct: float = _cct[0]
            self.duv: float = _cct[1]
            self.dominant_wl = float(dominant_wavelength(self.xy, [1 / 3, 1 / 3])[0])
            self.purity: float = colorimetric_purity(self.xy, (1 / 3, 1 / 3))  # type: ignore
            self.power: float = np.asarray(self.spd.values).sum()
            self.time = datetime.now().astimezone()

    def __str__(self) -> str:
        """
        Return printable string with interesting data for an observant reader.

        Returns
        -------
        str
        """
        return textwrap.dedent(
            f"""
            Spectral Measurement - {self.spectrometer_id}:
                time: {self.time}
                XYZ: {np.array2string(self.XYZ, formatter={'float_kind':lambda x: f"{x:.4f}"})}
                xy: {np.array2string(self.xy, formatter={'float_kind':lambda x: f"{x:.4f}"})}
                CCT: {self.cct:.0f} ± {self.duv:.5f}
                Dominant WL: {self.dominant_wl:.1f} @ {self.purity * 100:.1f}%
                Exposure: {self.exposure:.3f}
            """  # noqa: E501
        )

    def __repr__(self) -> str:
        """Return a loggable 1-line string for some basic details for this
        measurement.

        Returns
        -------
        str
        """

        return f"Spectral Measurement - {self.spectrometer_id}, Time: {self.time}, XYZ = {self.XYZ}"  # noqa: E501

    def __eq__(self, other: object) -> bool:
        """Check equality to another `specio.spectrometers.common.Measurement.` Based on
        multiple subfields.

        True if "exposure", "spectrometer_id", "XYZ", "xy", "dominant_wl",
        "power", "time", "cct", "duv" are all equal.

        Parameters
        ----------
        other : Measurement
            The object to compare to

        Returns
        -------
        bool
        """

        keys = [
            "spd",
            "exposure",
            "spectrometer_id",
            "XYZ",
            "xy",
            "dominant_wl",
            "purity",
            "power",
            "time",
            "cct",
            "duv",
        ]
        data = [getattr(self, k) == getattr(other, k) for k in keys]
        return all(np.all(d) for d in data)


class SpecRadiometer(ABC):
    """
    Base class for spectroradiometers. Defines the minimum interface needed
    by implementing classes to collect measurement values from a spectrometer.
    """

    @property
    @abstractmethod
    def serial_number(self) -> str:
        """The serial number or id for the specific connected spectrometer

        Returns
        -------
        str
        """

        ...

    @property
    @abstractmethod
    def manufacturer(self) -> str:
        """The device manufacturer name

        Returns
        -------
        str
        """

        ...

    @cached_property
    @abstractmethod
    def model(self) -> str:
        """The model name or identifier for the device

        Returns
        -------
        str
        """

        ...

    @abstractmethod
    def _raw_measure(self) -> RawSPDMeasurement:
        """Conduct remote triggering and RPC / serial interaction to collect
        the raw spectrometer data from the device.

        Returns
        -------
        RawMeasurement
            A simple dataclass with the required parameters to produce fully
            defined :class:`specio.spectrometers.common.Measurement`
        """

        ...

    @property
    def readable_id(self) -> str:
        """Human readable spectrometer id. Defaults to "<MODEL> - <SERIAL>" if
        not set by the implementing class.

        Returns
        -------
        str
        """
        return f"{self.model} - {self.serial_number}"

    def measure(self, repetitions: int = 1) -> SPDMeasurement:
        """Trigger and collect one measurement from the LED tile

        Returns
        -------
        Measurement
            A dataclass containing various convenience calculations from python
            colour-science and the raw SPD.
        """

        if repetitions < 1:
            ArgumentError("Repetitions must be greater than 1")

        _rm: list[RawSPDMeasurement] = []
        for i in range(repetitions):
            _rm += [self._raw_measure()]

        if len(_rm) == 1:
            return SPDMeasurement.FromRaw(_rm[0])

        spd_values = np.asarray([m.spd.values for m in _rm]).mean(axis=0)
        spd = SpectralDistribution(data=spd_values, domain=_rm[0].spd.domain)
        exposure = np.mean([m.exposure for m in _rm]).item()
        id = _rm[0].spectrometer_id

        return SPDMeasurement.FromRaw(
            RawSPDMeasurement(spd=spd, exposure=exposure, spectrometer_id=id)
        )


@final
class VirtualSpectrometer(SpecRadiometer):
    """
    Basic spectroradiometer interface. Implements a virtual spectrometer
    returning random colors for basic testing
    """

    def __init__(self):
        super().__init__()

    @property
    def manufacturer(self) -> str:
        """Return "specio" as the manufacturer of this virtual spectrometer

        Returns
        -------
        str
        """
        return "specio"

    @cached_property
    def model(self):
        """The model name or model signature from the spectrometer.

        Returns
        -------
        str
        """
        return "Virtual Random Spectrometer"

    @property
    def serial_number(self):
        """The serial number of the spectrometer

        Returns
        -------
        str
        """
        return "0000-0000"

    def _raw_measure(self) -> RawSPDMeasurement:
        """Return a random SPD generated by combining three gaussian spectra

        Returns
        -------
        RawMeasurement
            A simple dataclass with the required parameters to produce fully
            defined :class:`specio.spectrometers.common.Measurement`
        """
        peaks = np.random.randint([460, 510, 600], [480, 570, 690], 3)
        widths = np.random.randint(40, 80, 3)
        powers = np.random.randint(10, 40, 3) / 1000
        spd = sd_multi_leds(
            peak_wavelengths=peaks,
            half_spectral_widths=widths / 2,
            peak_power_ratios=powers,
        )
        spd.interpolator = SpragueInterpolator

        _measurement = RawSPDMeasurement(
            spd=spd,
            exposure=1,
            spectrometer_id="Virtual Spectrometer",
        )
        return _measurement



import argparse
import os
import sys
from pathlib import Path

from specio.common import get_valid_filename
from specio.serialization.csmf import (
    CSMF_Data,
    CSMF_Metadata,
    load_csmf_file,
    save_csmf_file,
)


def main(*args):
    parser = argparse.ArgumentParser()
    parser.add_argument("file", help="The csmf file to strip data from")

    parser.add_argument("-o", "--out-dir", help="Output directory", default=None)

    if args == ():
        args = None
    args = parser.parse_args(args)

    file_path = Path(args.file)
    file_data = load_csmf_file(file_path)

    output_path = file_path.parent if args.out_dir is None else Path(args.out_dir)

    file_data.metadata = CSMF_Metadata(software="specio:csmf_anonymize")

    output_path = output_path.joinpath(
        get_valid_filename(file_data.shortname)
    ).with_suffix(".csmf")
    output_path.parent.mkdir(parents=True, exist_ok=True)

    ml = CSMF_Data(
        test_colors=file_data.test_colors,
        measurements=file_data.measurements,
        metadata=file_data.metadata,
        order=file_data.order,
    )

    save_csmf_file(file=str(output_path), ml=ml)
    if args is None:
        output_path = output_path.relative_to(os.getcwd())
        sys.stdout.writelines(f"{output_path!s}")
    return output_path


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: specio/serialization/protobuf/measurements.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'specio/serialization/protobuf/measurements.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from specio.serialization.protobuf import common_pb2 as specio_dot_serialization_dot_protobuf_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0specio/serialization/protobuf/measurements.proto\x12\x0cspecio.proto\x1a*specio/serialization/protobuf/common.proto\"\xff\x01\n\x17\x43olorimeter_Measurement\x12$\n\x03XYZ\x18\x01 \x01(\x0b\x32\x17.specio.proto.XYZ_value\x12\"\n\x02xy\x18\x02 \x01(\x0b\x32\x16.specio.proto.xy_value\x12$\n\x03\x63\x63t\x18\x05 \x01(\x0b\x32\x17.specio.proto.cct_value\x12\x13\n\x0b\x64ominant_wl\x18\x06 \x01(\x01\x12\x0e\n\x06purity\x18\x08 \x01(\x01\x12%\n\x04time\x18\x07 \x01(\x0b\x32\x17.specio.proto.Timestamp\x12\x10\n\x08\x65xposure\x18\x10 \x01(\x01\x12\x16\n\x0e\x63olorimeter_id\x18\x11 \x01(\t\"\xbe\x02\n\x0fSPD_Measurement\x12/\n\x03spd\x18\x01 \x01(\x0b\x32\".specio.proto.SpectralDistribution\x12\x10\n\x08\x65xposure\x18\x02 \x01(\x01\x12$\n\x03XYZ\x18\x03 \x01(\x0b\x32\x17.specio.proto.XYZ_value\x12\"\n\x02xy\x18\x04 \x01(\x0b\x32\x16.specio.proto.xy_value\x12$\n\x03\x63\x63t\x18\x05 \x01(\x0b\x32\x17.specio.proto.cct_value\x12\x13\n\x0b\x64ominant_wl\x18\x06 \x01(\x01\x12\x0e\n\x06purity\x18\n \x01(\x01\x12\r\n\x05power\x18\x07 \x01(\x01\x12\x17\n\x0fspectrometer_id\x18\x08 \x01(\t\x12%\n\x04time\x18\t \x01(\x0b\x32\x17.specio.proto.TimestampJ\x04\x08\x10\x10\x11\"\xc9\x03\n\tCSFM_File\x12\x39\n\x0cmeasurements\x18\x04 \x03(\x0b\x32#.specio.proto.CSFM_File.Measurement\x12\x37\n\x10spd_measurements\x18\x01 \x03(\x0b\x32\x1d.specio.proto.SPD_Measurement\x12\x36\n\x0btest_colors\x18\x03 \x03(\x0b\x32!.specio.proto.CSFM_File.TestColor\x12\r\n\x05order\x18\x02 \x03(\x05\x12\r\n\x05notes\x18\x10 \x01(\t\x12\x0e\n\x06\x61uthor\x18\x11 \x01(\t\x12\x10\n\x08location\x18\x12 \x01(\t\x12\x10\n\x08software\x18\x13 \x01(\t\x12\x11\n\tancillary\x18\x14 \x01(\x0c\x1a\x87\x01\n\x0bMeasurement\x12/\n\x03spd\x18\x01 \x01(\x0b\x32\x1d.specio.proto.SPD_MeasurementH\x00\x88\x01\x01\x12\x37\n\x03xyz\x18\x02 \x01(\x0b\x32%.specio.proto.Colorimeter_MeasurementH\x01\x88\x01\x01\x42\x06\n\x04_spdB\x06\n\x04_xyz\x1a!\n\tTestColor\x12\t\n\x01\x63\x18\x01 \x03(\r\x12\t\n\x01\x66\x18\x02 \x03(\x02\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'specio.serialization.protobuf.measurements_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_COLORIMETER_MEASUREMENT']._serialized_start=111
  _globals['_COLORIMETER_MEASUREMENT']._serialized_end=366
  _globals['_SPD_MEASUREMENT']._serialized_start=369
  _globals['_SPD_MEASUREMENT']._serialized_end=687
  _globals['_CSFM_FILE']._serialized_start=690
  _globals['_CSFM_FILE']._serialized_end=1147
  _globals['_CSFM_FILE_MEASUREMENT']._serialized_start=977
  _globals['_CSFM_FILE_MEASUREMENT']._serialized_end=1112
  _globals['_CSFM_FILE_TESTCOLOR']._serialized_start=1114
  _globals['_CSFM_FILE_TESTCOLOR']._serialized_end=1147
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: specio/serialization/protobuf/common.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'specio/serialization/protobuf/common.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*specio/serialization/protobuf/common.proto\x12\x0cspecio.proto\"9\n\rSpectralShape\x12\r\n\x05start\x18\x01 \x01(\x01\x12\x0b\n\x03\x65nd\x18\x02 \x01(\x01\x12\x0c\n\x04step\x18\x03 \x01(\x01\"\x82\x01\n\x14SpectralDistribution\x12*\n\x05shape\x18\x01 \x01(\x0b\x32\x1b.specio.proto.SpectralShape\x12\x0e\n\x06values\x18\x02 \x03(\x01\x12\x12\n\nvalues_old\x18\x04 \x03(\x02\x12\x11\n\x04name\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\x07\n\x05_name\",\n\tXYZ_value\x12\t\n\x01X\x18\x01 \x01(\x01\x12\t\n\x01Y\x18\x02 \x01(\x01\x12\t\n\x01Z\x18\x03 \x01(\x01\" \n\x08xy_value\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\"%\n\tcct_value\x12\x0b\n\x03\x63\x63t\x18\x01 \x01(\x01\x12\x0b\n\x03\x64uv\x18\x02 \x01(\x01\"\x1c\n\tTimestamp\x12\x0f\n\x07timestr\x18\x01 \x01(\tb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'specio.serialization.protobuf.common_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SPECTRALSHAPE']._serialized_start=60
  _globals['_SPECTRALSHAPE']._serialized_end=117
  _globals['_SPECTRALDISTRIBUTION']._serialized_start=120
  _globals['_SPECTRALDISTRIBUTION']._serialized_end=250
  _globals['_XYZ_VALUE']._serialized_start=252
  _globals['_XYZ_VALUE']._serialized_end=296
  _globals['_XY_VALUE']._serialized_start=298
  _globals['_XY_VALUE']._serialized_end=330
  _globals['_CCT_VALUE']._serialized_start=332
  _globals['_CCT_VALUE']._serialized_end=369
  _globals['_TIMESTAMP']._serialized_start=371
  _globals['_TIMESTAMP']._serialized_end=399
# @@protoc_insertion_point(module_scope)



"""
Specio
======
Provides support for interacting with various hardware spectrometers.
"""

__version__ = "0.4.0a0"
__author__ = "Tucker Downs"
__copyright__ = "Copyright 2022 Specio Developers"
__license__ = "MIT License - https://github.com/tjdcs/specio/blob/main/LICENSE.md"
__maintainer__ = "Tucker Downs"
__email__ = "tucker@tjdcs.dev"
__status__ = "Development"


from .common.colorimeters import ColorimeterMeasurement, VirtualColorimeter
from .common.spectrometers import (
    SPDMeasurement,
    VirtualSpectrometer,
)
from .device_implementations import colorimetry_research, konica_minolta

__all__ = [
    "SPDMeasurement",
    "VirtualSpectrometer",
    "ColorimeterMeasurement",
    "VirtualColorimeter",
    "colorimetry_research",
    "konica_minolta",
]


def _config__specio_logger() -> None:
    """
    Configure the default specio logger "specio"
    """
    import logging

    log = logging.getLogger("specio")

    stream_handler = logging.StreamHandler()
    stream_handler.formatter = logging.Formatter(
        fmt="[%(asctime)s.%(msecs)d] %(levelname)s - %(name)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    log.addHandler(stream_handler)


_config__specio_logger()



"""
Define classes and functions for controlling Konica-Minolta brand spectrometers,
namely the CS2000.
"""

import platform
import struct
import time
from collections.abc import Mapping
from enum import Enum
from functools import cached_property
from textwrap import dedent
from types import MappingProxyType
from typing import Any, NamedTuple, cast, final

import aenum
import serial
from colour import SpectralDistribution, SpectralShape
from serial.tools import list_ports
from serial.tools.list_ports_common import ListPortInfo

from specio.common import RawSPDMeasurement, SpecRadiometer


class ResponseCode(aenum.MultiValueEnum):
    """Defines various CS2000 response codes and helpful error messages as a
    multivalue enum. Not all Enum instances have a second value. When
    present, the second value is a helpful message for users.

    The canonical value of each instance is the response code in `bytes`
    expected by the CS2000 serial protocol.
    """

    OK = b"OK00"

    MALFORMED_COMMAND = b"ER00", "The command string could not be interpreted"
    RESPONSE_NOT_READY = b"ER02", "Measurement response is not ready yet."
    OVER_EXPOSED = b"ER10", "Radiance too high for measurement mode"
    PARAMETER_ERROR = (
        b"ER17",
        "The command parameters could not be interpreted",
    )

    TEMPURTURE_ABNORMALITY_0 = b"ER51", "Device temperature abnormal. "
    TEMPURTURE_ABNORMALITY_1 = b"ER52", "Device temperature abnormal."

    SYNC_ERROR = b"ER71", "Could not synchronize to light source flicker"
    TARGET_ABNORMALITY = b"ER83", "Measurement Area Abnormality"

    RESERVED = b"ER--", "The response code could not be read by specio"

    @classmethod
    def _missing_(cls, _: object) -> Any:
        return cls.RESERVED


class CommandResponse(NamedTuple):
    """Internal Command Response representation for the CS2000 serial protocol"""

    code: ResponseCode
    data: tuple[bytes, ...]


class WriteCommandError(Exception):
    """Error thrown when there is an issue writing commands to a CS2000"""

    def __init__(self, message: str, cr: CommandResponse, *args: object) -> None:
        super().__init__(message, *args)
        self.command_response = cr


class SyncMode(bytes, Enum):
    NO_SYNC = b"0"
    INTERNAL = b"1"
    EXTERNAL = b"2"


class SyncSpeedSetting(NamedTuple):
    mode: SyncMode
    frequency: float | None


class InternalNDMode(bytes, Enum):
    OFF = b"0"
    ON = b"1"
    AUTO = b"2"


class SpeedMode(bytes, Enum):
    NORMAL = b"0"
    FAST = b"1"
    MULTI_NORMAL = b"2"
    MULTI_FAST = b"4"
    MANUAL = b"3"


class SpeedModeSetting:
    @staticmethod
    def from_command_response(cr: CommandResponse):
        blist = cr.data
        mode = SpeedMode(blist[0])
        if mode in (SpeedMode.NORMAL, SpeedMode.FAST):
            nd_mode = InternalNDMode(blist[1])
            time = None
        elif mode in (SpeedMode.MULTI_FAST, SpeedMode.MULTI_NORMAL):
            time = int(blist[1].decode())
            nd_mode = InternalNDMode(blist[2])
        elif mode is SpeedMode.MANUAL:
            time = float(blist[1].decode()) / 1e6
            nd_mode = InternalNDMode(blist[2])

        return SpeedModeSetting(mode=mode, nd_mode=nd_mode, integration_time=time)

    def __init__(
        self,
        mode: SpeedMode,
        nd_mode: InternalNDMode = InternalNDMode.AUTO,
        integration_time: float | int | None = None,
    ):
        self.mode = mode
        self.nd_mode = nd_mode

        if mode in (SpeedMode.MULTI_NORMAL, SpeedMode.MULTI_FAST):
            if (
                integration_time is None
                or integration_time < 2
                or integration_time > 16
            ):
                raise ValueError(
                    f"""SpeedMode.{mode.name} requires integration_time time
                    to be between 2 and 16 integer seconds."""
                )
            self.time: int = int(round(integration_time))

        elif mode is SpeedMode.MANUAL:
            if (
                integration_time is None
                or integration_time < 0.005
                or integration_time > 120
            ):
                raise ValueError(
                    f"""SpeedMode.{mode.name} requires manual_integration time
                    to be between .005 and 120 seconds."""
                )
            self.time = int(round(integration_time * 1e6))

    def __bytes__(self):
        cmd_bytes = bytearray(self.mode + b",")
        if self.mode in (SpeedMode.MULTI_FAST, SpeedMode.MULTI_NORMAL):
            cmd_bytes += f"{self.time:02d},".encode()
        elif self.mode is SpeedMode.MANUAL:
            cmd_bytes += f"{self.time:09d},".encode()
        cmd_bytes += self.nd_mode
        return bytes(cmd_bytes)


DEFAULT_SPEED_MODE_SETTING = SpeedModeSetting(
    SpeedMode.NORMAL, InternalNDMode.AUTO, integration_time=None
)
HIGH_ACCURACY_MODE_SETTING = SpeedModeSetting(
    SpeedMode.MULTI_NORMAL, InternalNDMode.AUTO, integration_time=16
)


@final
class CS2000(SpecRadiometer):
    CS2000_SERIAL_KWARGS: Mapping = MappingProxyType(
        {
            "baudrate": 115200,
            "bytesize": 8,
            "parity": "N",
            "rtscts": True,
            "timeout": 1,
        }
    )

    @classmethod
    def discover(cls) -> "CS2000":
        """Attempt automatic discovery of the CS2000 serial port and return the
        CS2000 object.

        Returns
        -------
        CS2000
            A successfully automatic CS2000 object.

        Raises
        ------
        serial.SerialException
            If no serial port can be automatically linked.
        """

        if platform.system() == "Darwin":
            possible_ports = list_ports.grep("CS-2000")
        else:
            raise NotImplementedError(
                f"CS2000 device discovery is not implemented for {platform.system()}"
            )

        for p in possible_ports:
            p = cast(ListPortInfo, p)  # Typing for `Serial` is wrong

            sp = serial.Serial(p.device, **cls.CS2000_SERIAL_KWARGS)
            sp.read_all()
            sp.write(b"RMTS,1\n")
            response = sp.readline()

            if response == b"OK00\n":
                return CS2000(sp)

            # Normally called during gc but it's better to be explicit here.
            sp.close()

        raise serial.SerialException(
            dedent(
                """Couldn't find serial port automatically. Make sure the device is
                plugged in. Try finding the serial port manually with
                serial.list_ports.comports() or resetting the device."""
            )
        )

    def __init__(self, port: serial.Serial | str):
        """Connect to the CS2000 device at the specified serial port. Provides
        spectral measurement functionality and various functions for specifying
        the behavior / speed / etc... of the CS2000.

        Parameters
        ----------
        port : serial.Serial | str
            The serial port that the CS2000 should be connected on.
        """
        if isinstance(port, str):
            port = serial.Serial(port, **self.CS2000_SERIAL_KWARGS)

        self._port = port
        self._clear_buffer()
        self._write_cmd("RMTS,1")
        (
            self._manufacturer,
            self._model,
            self._serial_number,
        ) = self._get_identity()

    def _get_identity(self) -> tuple[str, str, str]:
        """
        Write the identify command and parse the results.

        Returns
        -------
        str
            The manufacturer string
        str
            The model string, should be one of "CS-2000" or "CS-2000A"
        str
            The serial number
        """
        _, identity_data = self._write_cmd("IDDR")

        make = "Konica-Minolta"

        if int(identity_data[1]) == 1:
            model = "CS-2000"
        elif int(identity_data[1]) == 2:
            model = "CS-2000A"
        else:
            model = identity_data[0].decode()

        sn = identity_data[2].decode()

        return make, model, sn

    def _clear_buffer(self):
        self._port.read_all()

    def _write_cmd(self, cmd: str | bytes, time_out: float = 0) -> CommandResponse:
        if time_out > 0:
            old_timeout = self._port.timeout
            self._port.timeout = time_out

        encoded_cmd = bytearray(cmd.encode() if isinstance(cmd, str) else cmd)

        if encoded_cmd[-1] != b"\n":
            encoded_cmd = encoded_cmd + b"\n"

        self._port.write(encoded_cmd)
        response = self._port.readline()[:-1]
        response = response.split(b",")

        code = ResponseCode(response[0])
        data = response[1:]

        command_response = CommandResponse(code, tuple(data))

        if time_out > 0:
            self._port.timeout = old_timeout

        if code.value[0:2] != b"OK":
            additional_info = code.values[1] if len(code.values) >= 2 else code.value
            raise WriteCommandError(
                "There was an error with the CS2000 command. " + additional_info,
                command_response,
            )

        return command_response

    @property
    def serial_number(self) -> str:
        """The device serial number"""
        return self._serial_number

    @property
    def manufacturer(self) -> str:
        """
        The device manufacturer. Should be "Konica-Minolta"
        """
        return self._manufacturer

    @cached_property
    def model(self) -> str:
        """The model name"""
        return self._model

    @property
    def syncmode(self) -> SyncSpeedSetting:
        if hasattr(self, "_sync_speed_setting"):
            return self._sync_speed_setting

        _, data = self._write_cmd("SCMR")
        mode = SyncMode(data[0])
        frequency = float(data[1]) / 100 if mode is SyncMode.INTERNAL else None
        self._sync_speed_setting = SyncSpeedSetting(mode, frequency)

        return self._sync_speed_setting

    @syncmode.setter
    def syncmode(self, new_mode: SyncSpeedSetting):
        update_cmd_str = bytearray(b"SCMS," + new_mode.mode)
        if new_mode.mode is SyncMode.INTERNAL:
            if (
                new_mode.frequency is None
                or new_mode.frequency < 20
                or new_mode.frequency > 200
            ):
                raise ValueError(
                    "Sync mode Internal requires frequency between 20Hz and 200Hz"
                )

            update_cmd_str += b"," + f"{round(new_mode.frequency * 100):.0f}".encode()

        del self._sync_speed_setting
        self._write_cmd(bytes(update_cmd_str))

    @property
    def speedmode(self):
        if hasattr(self, "_cur_sms"):
            return self._cur_sms
        cr = self._write_cmd(b"SPMR")
        self._cur_sms = SpeedModeSetting.from_command_response(cr)
        return self._cur_sms

    @speedmode.setter
    def speedmode(self, cr: SpeedModeSetting):
        code, _ = self._write_cmd(b"SPMS," + bytes(cr))

    def _raw_measure(self) -> RawSPDMeasurement:
        # Additional timeout recommended by KM manual
        response = self._write_cmd("MEAS,1", time_out=10)
        wait_time = int(response.data[0])  # KM estimated measurement time
        time.sleep(wait_time + 0.1)

        while True:
            try:
                self._clear_buffer()
                _, conditions = self._write_cmd("MEDR,0,0,1")
                break
            except WriteCommandError as e:
                if e.command_response.code is ResponseCode.RESPONSE_NOT_READY:
                    time.sleep(0.5)
                    continue
                else:
                    raise e  # noqa: TRY201 Transparently re-raise original exception

        _, spd0 = self._write_cmd("MEDR,1,1,1")
        _, spd1 = self._write_cmd("MEDR,1,1,2")
        _, spd2 = self._write_cmd("MEDR,1,1,3")
        _, spd3 = self._write_cmd("MEDR,1,1,4")

        spd_data = [
            *[struct.unpack(">f", bytes.fromhex(d.decode()))[0] for d in spd0],
            *[struct.unpack(">f", bytes.fromhex(d.decode()))[0] for d in spd1],
            *[struct.unpack(">f", bytes.fromhex(d.decode()))[0] for d in spd2],
            *[struct.unpack(">f", bytes.fromhex(d.decode()))[0] for d in spd3],
        ]
        spd = SpectralDistribution(spd_data, SpectralShape(380, 780, 1))
        exposure = float(conditions[2]) * 1e-6

        return RawSPDMeasurement(
            spd=spd, exposure=exposure, spectrometer_id=self.readable_id
        )
